@@ -34,9 +34,9 @@ import tempfile
 import uuid
 import shutil
 import random
-
 import imagehash
 from PIL import Image
+from io import BytesIO
 
 # Configure the logger
 logger = logging.getLogger('')
@@ -444,40 +444,7 @@ class Clustimage():
 
         if len(self.results['feat'])==0:
             return None
-        # if (self.params['method'] is not None) and ('hash' in self.params['method']) and (metric!='hamming'):
-            # logger.warning('In case of method=[%s], metric="hamming" with linkage="complete" is recommended.' %(self.params['method']))
 
-        # If exact hash matches are required.
-        # if (self.params['method'] is not None) and ('hash' in self.params['method']) and self.params_hash['exact_hash']:
-        #     logger.info('Updating cluster-labels based on the [%s] with threshold=%g change.' %(self.params['method'], self.params_hash['threshold']))
-        #     labels = np.zeros(self.results['feat'].shape[0]) *np.nan
-
-        #     # Exact hash matches that appear to be significantly similar based on the hashes.
-        #     if not self.params_hash['exact_hash']:
-        #         # method='parametric'
-        #         model = distfit(method='quantile', bound='down', multtest=None, distr=['norm', 'expon', 'uniform', 'gamma', 't'], alpha=self.params_hash['threshold'])
-        #         # Fit theoretical distribution
-        #         _ = model.fit_transform(self.results['feat'])
-        #         # model.plot()
-        #     cl_label=0
-        #     for i in range(0, self.results['feat'].shape[0]):
-        #         # Retrieve all hashes that are similar with minimum value
-        #         Iloc = self.results['feat'][i, :]<=self.params_hash['threshold']
-        #         # Retrieve hashes based on probabilities
-        #         if not self.params_hash['exact_hash']:
-        #             Iloc_p = model.predict(self.results['feat'][i, :], verbose=0)['y_proba']<=self.params_hash['threshold']
-        #             Iloc = np.logical_or(Iloc, Iloc_p)
-        #         # If nan value is found, label it with new cluster label
-        #         if np.all(np.isnan(labels[Iloc])):
-        #             labels[Iloc]=cl_label
-        #             cl_label=cl_label +1
-        #         else:
-        #             tmplabels=labels[Iloc]
-        #             uilabels = np.unique(tmplabels[~np.isnan(tmplabels)])
-        #             labels[np.isin(labels, uilabels)]=uilabels[0]
-        #             labels[Iloc] = uilabels[0]
-        # else:
-            # Init
         ce = clusteval(cluster=cluster, evaluate=evaluate, metric=metric, linkage=linkage, min_clust=min_clust, max_clust=max_clust, verbose=3)
         # Fit
         if cluster_space=='low':
@@ -810,13 +777,17 @@ class Clustimage():
             filenames : list of str.
 
         """
+        # Make list of str
+        if isinstance(pathnames, str): pathnames=[pathnames]
+
         # Filter images on min-number of pixels in image
         min_nr_pixels = 8
+
         # Check file existence on disk
-        if isinstance(pathnames, str): pathnames=[pathnames]
         pathnames = list(np.array(pathnames)[list(map(os.path.isfile, pathnames))])
         filenames = list(map(basename, pathnames))
         idx = range(0, len(pathnames))
+
         # Output dict
         out = {'img': None, 'pathnames': pathnames, 'filenames': filenames}
 
@@ -827,14 +798,24 @@ class Clustimage():
             flatten=False
 
         # Read and preprocess data
-        img = list(map(lambda x: self.imread(x, colorscale=grayscale, dim=dim, flatten=flatten), tqdm(pathnames, disable=disable_tqdm())))
+        imgs = list(map(lambda x: self.imread(x, colorscale=grayscale, dim=dim, flatten=flatten, return_succes=True), tqdm(pathnames, disable=disable_tqdm())))
+        img, imgOK = zip(*imgs)
+        img = np.array(img)
 
         # Remove the images that could not be read
+        I_corrupt = np.array(imgOK)==False
+        if np.any(I_corrupt)>0:
+            logger.info("[%.0d] Corrupt image(s) removed.", (sum(I_corrupt)))
+            filenames = np.array(filenames)[~I_corrupt]
+            pathnames = np.array(pathnames)[~I_corrupt]
+            img = img[~I_corrupt]
+
+        # Remove the images that are too small
         if np.where(np.array(list(map(len, img)))<min_nr_pixels)[0]:
             logger.info("Images with < %.0d pixels are detected and excluded.", (min_nr_pixels))
 
         idx = np.where(np.array(list(map(len, img)))>=min_nr_pixels)[0]
-        img = np.array(img)[idx]
+        img = img[idx]
         if flatten: img = np.vstack(img)
 
         # Remove not readable images and return
@@ -993,6 +974,8 @@ class Clustimage():
         if isinstance(Xraw, list) or isinstance(Xraw[0], str):
             # Make sure that list in lists are flattend
             Xraw = np.hstack(Xraw)
+            # Check whether url and store all url images to tempdir on disk.
+            Xraw = url2disk(Xraw, self.params['tempdir'])
             # Do not store in the object if the find functionality is used
             X = self.preprocessing(Xraw, grayscale=self.params['cv2_imread_colorscale'], dim=self.params['dim'], flatten=flatten)
 
@@ -1321,7 +1304,7 @@ class Clustimage():
         # Return
         return out
 
-    def imread(self, filepath, colorscale=1, dim=(128, 128), flatten=True):
+    def imread(self, filepath, colorscale=1, dim=(128, 128), flatten=True, return_succes=False):
         """Read and pre-processing of images.
 
         Description
@@ -1354,7 +1337,7 @@ class Clustimage():
             Imported and processed image.
 
         Examples
-        ---------
+        --------
         >>> # Import libraries
         >>> from clustimage import Clustimage
         >>> import matplotlib.pyplot as plt
@@ -1377,6 +1360,7 @@ class Clustimage():
         """
         logger.debug('[%s]' %(filepath))
         img=[]
+        readOK = False
         try:
             # Read the image
             img = _imread(filepath, colorscale=colorscale)
@@ -1386,10 +1370,15 @@ class Clustimage():
             img = imresize(img, dim=dim)
             # Flatten the image
             if flatten: img = img_flatten(img)
+            # OK
+            readOK = True
         except:
             logger.warning('Could not read: [%s]' %(filepath))
         # Return
-        return img
+        if return_succes:
+            return img, readOK
+        else:
+            return img
 
     def save(self, filepath='clustimage.pkl', overwrite=False):
         """Save model in pickle file.
@@ -1590,7 +1579,7 @@ class Clustimage():
                 imagebox = offsetbox.AnnotationBbox(offsetbox.OffsetImage(img, cmap=cmap, zoom=zoom), xycoord[i, :])
                 ax.add_artist(imagebox)
 
-    def scatter(self, dotsize=15, legend=False, zoom=0.3, img_mean=True, text=True, plt_all=False, figsize=(15, 10)):
+    def scatter(self, dotsize=15, legend=False, zoom=0.3, img_mean=True, text=True, plt_all=False, density=False, figsize=(15, 10), args_scatter={}):
         """Plot the samples using a scatterplot.
 
         Parameters
@@ -1607,44 +1596,72 @@ class Clustimage():
             None : Do not plot the image.
         text : bool, (default: True)
             Plot the cluster labels.
+        density : bool, (default: Fale)
+            Plot the density over the clusters.
         figsize : tuple, (default: (15, 10).
             Size of the figure (height,width).
+        args_scatter : dict, default: {}.
+            Arguments for the scatter plot. The following are default:
+            {'title': '',
+            'fontsize': 18,
+            'fontcolor': [0, 0, 0],
+            'xlabel': 'x-axis',
+            'ylabel': 'y-axis',
+            'cmap': cmap,
+            'density': False,
+            'gradient': None,
+            }
 
         Returns
         -------
-        None.
+        tuple (fig, ax)
+
+        Examples
+        --------
+        >>> from clustimage import Clustimage
+        >>>
+        >>> cl = Clustimage()
+        >>> X = cl.import_example(data='mnist')
+        >>> # Run the model to find the optimal clusters.
+        >>> results = cl.fit_transform(X)
+        >>> Make scatter plots
+        >>> cl.scatter()
+        >>> # More input arguments for the scatterplot
+        >>> cl.scatter(dotsize=35, args_scatter={'fontsize':24, 'density':'#FFFFFF', 'cmap':'Set2'})
 
         """
         # Check status
+        fig, ax = None, None
         self._check_status()
         if self.results['xycoord'] is None:
             logger.warning('Missing x,y coordinates in results dict. Hint: try to first run: cl.embedding(Xfeat)')
             return None
         # Set default settings
-        cmap = plt.cm.gray if self.params['grayscale'] else None
+        # if args_scatter.get('cmap', None) is None:
+        cmap = plt.cm.gray if self.params['grayscale'] else 'Set1'
         # Get the cluster labels
         labels = self.results.get('labels', None)
         if labels is None: labels=np.zeros_like(self.results['xycoord'][:, 0]).astype(int)
 
-        if text:
-            text_labels=labels
-        else:
-            text_labels=None
-
         # Make scatterplot
-        if self.params.get('cluster_space', None) is None:
-            title = ('tSNE plot.')
-            logger.warning('Run .fit_transform() or .cluster() to colour based on the samples on the cluster labels.')
-        else:
-            title = ('tSNE plot. Samples are coloured on the cluster labels (%s dimensional).' %(self.params['cluster_space']))
+        if args_scatter.get('title', None) is None:
+            if self.params.get('cluster_space', None) is None:
+                title = ('tSNE plot.')
+                logger.warning('Run .fit_transform() or .cluster() to colour based on the samples on the cluster labels.')
+            else:
+                title = ('tSNE plot. Samples are coloured on the cluster labels (%s dimensional).' %(self.params['cluster_space']))
+
+        # Defaults
+        default_scatter = {'title': title, 'fontsize': 18, 'fontcolor': [0, 0, 0], 'xlabel': 'x-axis', 'ylabel': 'y-axis', 'cmap': 'Set1', 'density': False, 'gradient': None, 'figsize': figsize, 'legend': True, 's': dotsize}
+        args_scatter = {**default_scatter, **args_scatter}
 
         # Set logger to warning-error only
         verbose = logger.getEffectiveLevel()
         set_logger(verbose=40)
 
         # Scatter
-        colours=np.vstack(colourmap.fromlist(labels)[0])
-        fig, ax = scatterd(self.results['xycoord'][:, 0], self.results['xycoord'][:, 1], s=dotsize, c=colours, label=text_labels, figsize=figsize, title=title, fontsize=18, fontcolor=[0, 0, 0], xlabel='x-axis', ylabel='y-axis')
+        # colours=np.vstack(colourmap.fromlist(labels)[0])
+        fig, ax = scatterd(self.results['xycoord'][:, 0], self.results['xycoord'][:, 1], labels=labels, **args_scatter)
 
         if hasattr(self, 'results_unique'):
             if img_mean:
@@ -1682,6 +1699,8 @@ class Clustimage():
 
         # Restore verbose status
         set_logger(verbose=verbose)
+        # Return
+        return fig, ax
 
     def plot_unique(self, cmap=None, img_mean=True, show_hog=False, figsize=(15, 10)):
         """Plot unique images.
@@ -2048,8 +2067,6 @@ def store_to_disk(Xraw, dim, tempdir):
     return pathnames, filenames
 
 # %% Unique without sort
-
-
 def unique_no_sort(x):
     """Unique without sort."""
     x = x[x!=None]
@@ -2374,7 +2391,8 @@ def get_params_hash(hashmethod, params_hash):
     elif hashmethod == 'whash-haar':
         hashfunc = imagehash.whash
     elif hashmethod == 'whash-db4':
-        def hashfunc(img): return imagehash.whash(img, mode='db4')
+        def hashfunc(img):
+            return imagehash.whash(img, mode='db4')
     elif hashmethod == 'colorhash':
         hashfunc = imagehash.colorhash
     elif hashmethod == 'crop-resistant':
@@ -2393,6 +2411,81 @@ def get_params_hash(hashmethod, params_hash):
     return params_hash
 
 
+# %%
+def url2disk(urls, save_dir):
+    """Write url locations to disk.
+
+    Description
+    -----------
+    Import images from url locations and store to disk.
+
+    Parameters
+    ----------
+    urls : list
+        list of url locations with image path.
+    save_dir : str
+        location to disk.
+
+    Returns
+    -------
+    urls : list of str.
+        list to url locations that are now stored on disk.
+
+    Examples
+    --------
+    >>> # Init with default settings
+    >>> model = Undouble()
+    >>>
+    >>> # Importing the files files from disk, cleaning and pre-processing
+    >>> url_to_images = ['https://erdogant.github.io/datasets/images/flower_images/flower_orange.png',
+    >>>                  'https://erdogant.github.io/datasets/images/flower_images/flower_white_1.png',
+    >>>                  'https://erdogant.github.io/datasets/images/flower_images/flower_white_2.png',
+    >>>                  'https://erdogant.github.io/datasets/images/flower_images/flower_yellow_1.png',
+    >>>                  'https://erdogant.github.io/datasets/images/flower_images/flower_yellow_2.png']
+    >>>
+    >>> # Import into model
+    >>> model.import_data(url_to_images)
+    >>>
+    >>> # Compute image-hash
+    >>> model.compute_hash(method='phash', hash_size=16)
+    >>>
+    >>> # Find images with image-hash <= threshold
+    >>> model.group(threshold=0)
+    >>>
+    >>> # Plot the images
+    >>> model.plot()
+
+    """
+    idx_url = np.where(list(map(lambda x: x[0:4]=='http', urls)))[0]
+    if len(idx_url)>0:
+        logger.info('[%.0d] urls are detected and stored on disk: [%s]' %(len(idx_url), save_dir))
+
+    if os.path.isdir(save_dir):
+        for idx in idx_url:
+            try:
+                # Make connection to file
+                response = requests.get(urls[idx])
+                img = Image.open(BytesIO(response.content))
+                # Get url
+                url = urlparse(urls[idx])
+                # Extract filename from url
+                url_filename = os.path.basename(url.path)
+                path_to_file = os.path.join(save_dir, url_filename)
+                if os.path.isfile(path_to_file):
+                    logger.info('[%s] already exists and is overwritten.' %(url_filename))
+                # save a image using extension
+                img.save(path_to_file)
+                # Store new location
+                urls[idx] = path_to_file
+            except FileNotFoundError:
+                logger.warning('error downloading file from [%s]' %(urls[idx]))
+        else:
+            logger.warning('[%s] does not exits.' %(save_dir))
+
+    # Return
+    return urls
+
+
 # %% Main
 # if __name__ == "__main__":
 #     import clustimage as clustimage
