@@ -1,9 +1,41 @@
 # %%
-# import clustimage
+import clustimage
 # print(dir(clustimage))
 # print(clustimage.__version__)
 
 
+# %% Import list of images from url adresses
+from clustimage import Clustimage
+
+# Initialize
+cl = Clustimage(method='pca', embedding='umap')
+
+# Importing the files files from disk, cleaning and pre-processing
+url_to_images = ['https://erdogant.github.io/datasets/images/flower_images/flower_orange.png',
+                 'https://erdogant.github.io/datasets/images/flower_images/flower_white_1.png',
+                 'https://erdogant.github.io/datasets/images/flower_images/flower_white_2.png',
+                 'https://erdogant.github.io/datasets/images/flower_images/flower_yellow_1.png',
+                 'https://erdogant.github.io/datasets/images/flower_images/flower_yellow_2.png']
+
+# Import into model
+X = cl.import_data(url_to_images)
+
+# Extract features using method
+Xfeat = cl.extract_feat(X)
+
+# Embedding
+xycoord = cl.embedding(Xfeat)
+
+# Cluster
+labels = cl.cluster()
+
+# Return
+results = cl.results
+
+# Make plots
+cl.plot()
+cl.dendrogram()
+
 # %%
 from clustimage import Clustimage
 
@@ -11,8 +43,8 @@ from clustimage import Clustimage
 cl = Clustimage(method='pca', embedding='umap')
 
 # Import data
-# Xraw = cl.import_example(data='flowers')
-Xraw = cl.import_example(data='mnist')
+Xraw = cl.import_example(data='flowers')
+# Xraw = cl.import_example(data='mnist')
 # Xraw = cl.import_example(data='faces')
 
 # Import data in a standardized manner
@@ -35,8 +67,7 @@ results = cl.results
 
 # Plots
 cl.clusteval.plot()
-cl.scatter()
-cl.scatter(zoom=None, img_mean=False, dotsize=20, text=False, plt_all=True)
+cl.scatter(zoom=None, dotsize=200, figsize=(25, 15), args_scatter={'fontsize':24, 'gradient':'#FFFFFF', 'cmap':'Set2', 'legend':True})
 cl.plot_unique()
 cl.plot()
 cl.dendrogram()
